/*
 * WebMap.java
 *
 * Created on December 28, 2002, 11:16 AM
 *
 * @author  Richard Kingston
 * @version 0.1
 */
package uk.ac.leeds.ccg.ppgis;

import java.io.*;
import java.net.*;
import java.awt.*;
import java.awt.image.*;
import java.awt.event.*;
import java.applet.*;

import uk.ac.leeds.ccg.geotools.*;
import uk.ac.leeds.ccg.widgets.*;

/*
 * This is the main class for the Applet. It displays shapefiles and allows users
 * to query the map and then make comments about map features.
 */

public class WebMap extends Applet implements ActionListener, 
    ViewerClickedListener, MouseListener {
    
    Viewer mapViewer = new Viewer();
    ShapefileReader sfr;
    Theme theme[];
    HighlightManager hlm[];
    GeoLabel geoLabel;
    GeoPoint location;
    
    Font titleFont = new Font("Arial", Font.BOLD, 14);
    Font labelFont = new Font("Arial", Font.BOLD, 12);
    Color darkRed = new Color(204,0,0);
    Color gtBlue = new Color(57,24,198);
    Color gtGreen = new Color(132,173,156);
    
    // External URL problems
    boolean badURL;
    
    // Parameter objects
    int NumOfLayers;
    String processDefaults;
    String defaultString;
    
    // ShapeFileReader progress box
    Progress info;
    
    // About VS pop-up window
    AboutVS avs;
    
    // Image loading
    Panel imagePanel = new Panel();
    Image gtImage;
    
    // Users comment text input
    String userInput;
    String pickedGeoLabel;
    
    /**
     * Put GeoTools image into Applet
     */
    public void paint(Graphics g) {
        g.drawImage(gtImage, 0, 0, 150, 51, this);
    }
    
    /**
     * For some reason GeoTools Themes in the init() method won't appear
     * in the browser window but will do it in start() method ?!?
     */
    public void start(){
        for (int j = 0; j < NumOfLayers; j++) {
            mapViewer.addTheme(theme[j]);
        }
    }
    
    /**
     * Main part of the Applet which loads data and sets
     * the layout of the screen
     */
    public void init() {
        // Load images
        gtImage = getImage(getDocumentBase(), "geotools.gif");
        
        // Create a Panel for the maps to sit on.
        mapArea = new Panel();
        mapArea.setBackground(gtGreen);
        
        // Parameter objects
        String shadeBy;
        String labelBy;
        String file;
        String lut;
        String none;
        String missing;
        
        // Geotools objects
        Shader shader;
        repaint();
        URL base = getCodeBase();
        
        // Set up sfr progress box
        Frame f = new Frame();
        info = new Progress(f, false);
        info.setTitle("File Reading Progress");
        Label status = new Label("Starting - Please wait!");
        info.add(status, "Center");
        info.setLocationRelativeTo(this);
        info.show();
        status.setForeground(gtBlue);
        
        // Setup map layout
        setLayout(new BorderLayout());
        mapViewer.setBackground(Color.white);
        mapViewer.setSize(550,550);
        
        // Get some Parameters
        NumOfLayers = Integer.parseInt(getParameter("NumOfLayers"));
        processDefaults = getParameter("processDefaults");
        defaultString = getParameter("defaultString");
        
        // Create the toolbar
        ToolBar tools = new ToolBar(mapViewer);
        controls = new Panel();
        controls.add(tools);
        
        // Setup GeoLabel tips
        geoLabel = new GeoLabel();
        geoLabel.setForeground(darkRed);
        
        // Setup a file reader system to log user comments
        // This has not been finished as we no longer want to 
        // log users comments anymore
        /*{ try
          {
              String dirName = "/feedback_logs";
              String fileName = "slawit_comments.txt";
              File output = new File(dirName, fileName);
              output.createNewFile();
          }
          catch(IOException e)
          {
              System.out.println("Error writing log file " + e);
          }
        }*/
        
        // Hide GeoLabels if html parameter is set to false
        if (!((getParameter("showLabel")).equals("false")));
        
        // Setup holder for themes and max possible HighLightManagers.
        theme = new Theme[NumOfLayers];
        hlm = new HighlightManager[NumOfLayers];
        
        // Loop through number of shapefiles
        for (int i = 0; i < NumOfLayers; i++) {
            
            // Get Parameters for each shapefile
            shadeBy = getParameter("shadeBy" + i);
            labelBy = getParameter("labelBy" + i);
            file = getParameter("shapefile" + i);
            lut = getParameter("lut" + i);
            none = getParameter("missingColor" + i);
            missing = getParameter("missingValue" + i);
            
            // Update sfr progress box
            status.setText("Loading: " + file + " data.");
            status.validate();
            
            Color missingColor = (none == null) ? null : Color.decode(none);
            int missingValue = (missing == null) ? -9999 : (new Integer(missing)).intValue();
            
            try {
                sfr = new ShapefileReader(new URL(base,file));
            }
            catch (Exception e) {System.out.println(base + file + ".shp");}
            
            shader = null;
            
            try{
                shader = new DiscreteShader(new URL(base,lut));
            }
            catch(IOException e){
                System.err.println("Unable to load lutfile " + lut + 
                ": switching to mono shade of grey");
                shader = new MonoShader(Color.gray);
            }
            
            shader.setMissingValueCode(missingValue);
            if(missingColor != null) {
                shader.setMissingValueColor(missingColor);
            }
            
            // Add HighLightManager if we want to layer searchable (Could ideally
            // do with keeping the HighLightManager and turning off the visual effect)
            if (!(getParameter("light" + i).equals("false"))) {
                hlm[i] = new HighlightManager();
                theme[i] = sfr.getTheme(shader,shadeBy);
                theme[i].setHighlightManager(hlm[i]);
                geoLabel.addHighlightManager(hlm[i],sfr.readData(labelBy));
            }
            else {
                theme[i] = sfr.getTheme(shader,shadeBy);
            }
            
            // Add the mapViewer to the mapArea panel.
            mapArea.add(mapViewer);
            //  Add a listener to the map to get the real world coordinates
            mapViewer.addViewerClickedListener(this);
            mapViewer.addMouseListener(this);
            // Get rid of the sfr progress box
            info.setVisible(false);
        }
        
        // This where the GUI objects are created for the Applet
        // Main title text
        title1 = new Label("Virtual Slaithwaite Participatory Planning System");
        title1.setFont(titleFont);
        title1.setForeground(darkRed);
        
        // Panel for geolabel text
        labelPanel = new Panel();
        attributeLabel = new Label("You have selected:");
        attributeLabel.setFont(labelFont);
        attributeLabel.setForeground(gtBlue);
        labelPanel.add(attributeLabel);
        labelPanel.add(geoLabel);
        
        // Create a title
        title2 = new Label("User Instructions");
        title2.setFont(labelFont);
        title2.setForeground(gtBlue);
        
        // Create some instructions
        instructions = new TextArea(
        "The right hand part of the screen will show a map of the village " +
        "indicating the location of buildings, roads, the railway line, " +
        "river, canal and reservoir. You can obtain information about " +
        "features on the map by moving the mouse over them.\n" +
        "\n" +
        "To make comments you need to \"double\" click with the mouse " +
        "button and type your comments in to the box provided below.\n" +
        "\n" +
        "At the bottom of the map three navigation options provided.\n" +
        "\n" +
        "\"Reset:\" Select this button to re-draw the whole map.\n" +
        "\n" +
        "\"Zoom:\" Using the mouse select an area of the map you wish to zoom " +
        "into by pointing at the corner of the area to enlarge, holding " +
        "down the left button draw a square across the map - by moving " +
        "the mouse to the opposite corner of the area you wish to enlarge " +
        "and then release the mouse button. After a few seconds the map " +
        "will re-draw.\n" +
        "\n" +
        "\"Pan:\" After zooming in to an area you can move the map in " +
        "various directions. Click on the \'Pan\' button and hold " +
        "down the left mouse button and drag it across the map. " +
        "The map will move(pan) and redraw.", 12, 30, TextArea.SCROLLBARS_VERTICAL_ONLY);
        instructions.setEditable(false);
        
        // Create a title
        title3 = new Label("User Communication");
        title3.setFont(labelFont);
        title3.setForeground(gtBlue);
        
        // Attribute information
        attribute = new TextArea("", 2 ,30, TextArea.SCROLLBARS_VERTICAL_ONLY);
        attribute.setEditable(false);
        attribute.setEnabled(false);
        attribute.addMouseListener(this);
        
        // This textArea is for users to type in there coments.
        // This text is cleared away once the user click on the map
        comments = new TextArea("", 12, 30, TextArea.SCROLLBARS_VERTICAL_ONLY);
        comments.setText("You must select something on the map which you " +
        "wish to make a comment about before you can type in this box.");
        comments.setEditable(true);
        comments.setEnabled(false);
        comments.addMouseListener(new UserCommentsInput(userInput));
        
        // The 'send' button
        ok = new Button("Send your comments");
        ok.setEnabled(false);
        ok.addMouseListener(new UserCommentsInput(userInput));
        
        // CCG web page button
        ccg = new Button("System Development");
        ccg.setActionCommand("http://www.ccg.leeds.ac.uk/democracy/");
        ccg.addActionListener(this);
        ccg.setFont(labelFont);
        ccg.setForeground(gtBlue);
        
        // Pop-up About VS button
        about = new Button("About Virtual Slaithwaite");
        about.setEnabled(true);
        about.addActionListener(this);
        about.setFont(labelFont);
        about.setForeground(gtBlue);
        
        // Set-up a GridBagLayout manager for the main components of the Applet
        GridBagLayout gbl = new GridBagLayout();
        setLayout(gbl);
        GridBagConstraints cons = new GridBagConstraints();
        
        // Instruction title  cell
        cons.gridx = 0;
        cons.gridy = 2;
        cons.gridwidth = 1;
        cons.gridheight = 1;
        gbl.setConstraints(title2, cons);
        add(title2);
        
        // TextArea 1 cell
        cons.gridx = 0;
        cons.gridy = 3;
        cons.gridwidth = 2;
        cons.gridheight = 2;
        gbl.setConstraints(instructions, cons);
        add(instructions);
        
        // Comments title cell
        cons.gridx = 0;
        cons.gridy = 5;
        cons.gridwidth = 1;
        cons.gridheight = 1;
        gbl.setConstraints(title3, cons);
        add(title3);
        
        // Attribute details cell
        cons.gridx = 0;
        cons.gridy = 6;
        cons.gridwidth = 2;
        cons.gridheight = 1;
        gbl.setConstraints(attribute, cons);
        add(attribute);
        
        // User comments cell
        cons.gridx = 0;
        cons.gridy = 7;
        cons.gridwidth = 2;
        cons.gridheight = 1;
        gbl.setConstraints(comments, cons);
        add(comments);
        
        // Attribute details cell
        cons.gridx = 0;
        cons.gridy = 8;
        cons.gridwidth = 2;
        cons.gridheight = 1;
        cons.fill = GridBagConstraints.CENTER;
        gbl.setConstraints(ok, cons);
        add(ok);
        
        // About VS cell
        cons.gridx = 0;
        cons.gridy = 9;
        cons.gridwidth = 1;
        cons.gridheight = 1;
        gbl.setConstraints(about, cons);
        add(about);
        
        // Main title cell
        cons.gridx = 2;
        cons.gridy = 0;
        cons.gridwidth = 2;
        cons.fill = GridBagConstraints.HORIZONTAL;
        gbl.setConstraints(title1, cons);
        add(title1);
        
        // GeoLabel cell
        cons.gridx = 2;
        cons.gridy = 1;
        cons.gridwidth = 4;
        cons.gridheight = 1;
        cons.fill = GridBagConstraints.HORIZONTAL;
        gbl.setConstraints(labelPanel, cons);
        add(labelPanel);
        
        // MapArea cell
        cons.gridx = 2;
        cons.gridy = 2;
        cons.gridwidth = 4;
        cons.gridheight = 7;
        gbl.setConstraints(mapArea, cons);
        add(mapArea);
        
        // External URL link cell
        cons.gridx = 2;
        cons.gridy = 9;
        cons.gridwidth = 1;
        cons.gridheight = 1;
        gbl.setConstraints(ccg, cons);
        add(ccg);
        
        // Navigation cell
        cons.gridx = 3;
        cons.gridy = 9;
        cons.gridwidth = 3;
        cons.gridheight = 1;
        cons.fill = GridBagConstraints.HORIZONTAL;
        gbl.setConstraints(controls, cons);
        add(controls);
    }
    
    /**
     * This class resizes and positions the little progress box
     * that is thrown up while files are loading.
     */
    class Progress extends Dialog {
        Progress(Frame parent, boolean bool) {
            super(parent, bool);
            setSize(300, 100);
            setBounds(300,300,300,100);
        }
    }
    
    /**
     *  Method to get the realworld location of a mouse click on the viewer.
     *  This is useful as it allows coordinates to be attached to a users
     *  comments for future analysis.
     */
    public void viewerClicked(ViewerClickedEvent vce) {
        location = mapViewer.getMapGeoPoint();
    }
    
    /**
     *  Method which inserts the users choice above the comment box
     *  and then enables the comments box to allow the user to add
     *  their comment
     */
    public void mouseClicked(MouseEvent me){
        if (me.getClickCount() >= 1) {
            attribute.setText("You have selected: " + geoLabel.getText() +
            ". If you want to make a comment type it in box below.");
            // Store the current geoLabel
            pickedGeoLabel = geoLabel.getText();
            attribute.setEnabled(false);
            comments.setEnabled(true);
            comments.setText("");
        }
    }
    
    public void mousePressed(MouseEvent me) {
    }
    public void mouseReleased(MouseEvent me) {
    }
    public void mouseEntered(MouseEvent me) {
    }
    public void mouseExited(MouseEvent me) {
    }
    
    // Inner class to deal with user comments input and processing
    class UserCommentsInput implements MouseListener {
        public UserCommentsInput(String userInput) {
        }
        
        public void mouseClicked(MouseEvent e) {
            if (e.getSource() == comments) {
                ok.setEnabled(true);
            }
            if (e.getSource() == ok) {
                attribute.setText("Thank you! Make another comment " +
                "if you wish");
                attribute.setEditable(false);
                // Put users comments into a string called feedback and
                // deliminate with the | symbol. This also includes the
                // real world coordinates of the users mousde click
                String feedback = ("Location: " + "|" + location + "|" +
                " GeoLabel: " + "|" + pickedGeoLabel  + "|" +
                " User input: " + "|" + comments.getText());
                System.out.println(feedback);
                comments.setText("");
            }
        }
        
        public void mouseEntered(MouseEvent e) {
        }
        public void mouseExited(MouseEvent e) {
        }
        public void mousePressed(MouseEvent e) {
        }
        public void mouseReleased(MouseEvent e) {
        }
    }
    
    public void actionPerformed(ActionEvent ae) {
        // About VS dialog action
        if (ae.getSource() == about) {
            avs = new AboutVS(new Frame());
            avs.setLocationRelativeTo(this);
            avs.setVisible(true);
        }
        // External URL link from ccg Button
        String urlLink = ae.getActionCommand();
        try {
            URL url = new URL(urlLink);
            AppletContext context = getAppletContext();
            // Open the URL in a new web browser window
            context.showDocument(url, "_blank");
        }
        catch (MalformedURLException urlException) {
            badURL = true;
            System.out.println("There is a problem opening the web page: " +
            urlException);
            repaint();
        }
    }
    
    private Panel mapArea, controls, labelPanel;
    private TextArea instructions, attribute, comments;
    private Label title1, title2, title3, usersLabel, attributeLabel;
    private Button about, ccg, ok;
}